from os import path
from requests import get
from shutil import rmtree
from unicodedata import normalize
from urllib import urlretrieve, unquote_plus, unquote, quote
from xbmcaddon import Addon
from xbmcplugin import endOfDirectory, addDirectoryItem
from xbmcgui import ListItem, Dialog
from xbmcvfs import exists, mkdirs
from xbmc import translatePath, executebuiltin, getInfoLabel, executeJSONRPC, Player, sleep, log, getCondVisibility
from re import sub, match, search
import sys
import requests
import json
import codecs

reload(sys)
sys.setdefaultencoding('utf8')

__addon__ = Addon()
__scriptid__ = __addon__.getAddonInfo('id')
__version__ = __addon__.getAddonInfo('version')
__temp__ = unicode(translatePath(__addon__.getAddonInfo('profile')), 'utf-8')
__subs__ = unicode(translatePath(path.join(__temp__, 'subs')), 'utf-8')
__scriptname__ = __addon__.getAddonInfo('name')
__language__ = __addon__.getLocalizedString

# A few design and flow notes:
# First, the base code was taken from wizdom subs (and some ideas from opensubtitles). Eventually, as the ScrewZira
# API was completely different, I had to rewrite and re-adapt 90% of the code.
# Second, the way I designed the use flow (from UX POV) is for users that use Kodi mainly as Media Streamer from a
# local source (i.e. some accessible file path rather than an add-on streaming) and my intentions were that the user
# will always be served with the best match, preferably a non-negotiable one (single match).
#
# From Kodi service flow, the idea of how a subtitle service works is quite simple once you get it:
# 1. The user performs a 'search' action in which we list all possible subtitles (see get_subtitle_listing).
# 2. When listing them to the user, we already "set" a call for the 'download' action (i.e. what Kodi should do once
#    the user double-clicked the item, sort of onClick / onDownload listener).
# 3. Once a subtitle is selected, the service will be invoked with a 'download' action (recall it's a different context
#    than the 'search' action) and we'll try to fetch the specific ID, saving it to some temp so Kodi can take it and
#    place with the proper file name / stream context.
#
# As i'm no Python expert, I've tried to follow the recommended coding style of PEP8:
# https://www.python.org/dev/peps/pep-0008
# The code presumes Python 3+ syntax and packages (that's why I strongly recommend on using only Kodi Krypton 17+).
#
# You want extra features? Have some questions? I can be reached @ dev.cb.corp@gmail.com

def normalize_string(str):
    return normalize('NFKD', unicode(unicode(str, 'utf-8'))).encode('utf-8', 'ignore')


def download(id, subs_name):

    try:
        rmtree(__subs__)
    except:
        pass
    mkdirs(__subs__)

    dbg_log("download: label: %s, id: %s" % (subs_name, id))

    headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}
    url = "http://api.screwzira.com/Download"
    data_dl = {"request": {"subtitleID": id}}
    response = requests.post(url, data=json.dumps(data_dl), headers=headers)

    # For some obscured reason, I've noticed several subtitles data fails upon posting so we must notify the user to
    # try some other mean
    if not response.ok:
        Dialog().ok("Error D/L subtitle", "Can't get %s.\nTry manual search!" % subs_name, "Sorry :(")
        return ""

    # if len(response.text) > 20:
    #     dbg_log("download: data: "+response.text[:100])

    subs_file_name = path.join(__subs__, subs_name + '.srt')
    subs_file = codecs.open(subs_file_name, 'w', 'utf-8')
    subs_file.write(response.text)
    subs_file.close()
    return subs_file_name

def get_params(arg):
    """ Return the service command argument in an unordered tuples of <key,value> """
    param = []
    if len(arg) >= 2:
        params_ = arg
        cleaned_params = params_.replace('?', '')
        pairs_of_params = cleaned_params.split('&')
        param = {}
        for i in range(len(pairs_of_params)):
            split_params = pairs_of_params[i].split('=')
            if len(split_params) == 2:
                param[split_params[0]] = split_params[1]
    return param


def get_param(name, params):
    try:
        return unquote_plus(params[name])
    except:
        pass

def get_title_name(title):
    """  Try to extract from a video file name its leading title """
    title = title.lower()
    title_tokens = title.replace('_', ' ').replace('.', ' ').replace('-', ' ').split()
    # This is a mere hueristics but most torrent "groups" will follow this basic convention.
    title_name = ""
    res_list = {"480p", "720p", "1080p", "2160p"}   # Most popular resolutions for common titles
    for str_token in title_tokens:
        if str_token in res_list:
            break
        try:
            # ScrewZira API doesn't handle well years so it's yet another show stopper for us
            test_year = int(str_token)
            if 1900 <= test_year <= 2050:
                break
        except:
            pass
        if match(r"s\d+e\d+", str_token):
            break
        title_name += str_token + " "
    return title_name


def get_title_match_score(title, subs_name):
    """ Given a title, compare it in parts to a subtitle name and return a count stating the no. of matching tokens """
    if subs_name[0] == '\u200f':
        print("here")
    title = title.lower()
    title_tokens = title.replace('_', ' ').replace('.', ' ').replace('-', ' ').split()
    subs_name = subs_name.lower()
    subs_tokens = subs_name.replace('_', ' ').replace('.', ' ').replace('-', ' ').replace('\u200f', ' ').split()
    score = 0
    for str_token in title_tokens:
        if str_token in subs_tokens:
            score += 1
    return score


def get_subtitle_listing(label, imdb_id="", season=0, episode=0):
    """ Does all the neat work of getting and listing all found subtitles """

    headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}

    # Some publishers like to concatenate their "group" to the end so remove it for better subtitle matching
    # e.g. Best.Movie.Ever.2020[eztv] --> Best.Movie.Ever.2020
    if label.find('[') != -1:
        label = label.split('[')[0]

    # First, we try to find the best match and list it
    if len(imdb_id) > 3:
        # Though IMDB is our best candidate for an exact match, in recent years it seems most Kodi users doesn't use the
        # too complex IMDB info obtainers so I wouldn't presume this flow is likely to be used (and surely in non of my
        # empirical tests).
        url = "http://api.screwzira.com/FindFilm"
        json_req = {"request": {"SearchPhrase": imdb_id, "SearchType": "ImdbID", "Version": "1.0"}}
    elif season != 0:
        url = "http://api.screwzira.com/FindSeries"
        json_req = {"request": {"SearchPhrase": label, "SearchType": "Subtitle", "Season": str(season), "Episode":
            str(episode), "Version": "1.0"}}
    else:
        url = "http://api.screwzira.com/FindFilm"
        json_req = {"request": {"SearchPhrase": label, "SearchType": "Subtitle", "Version": "1.0"}}
    response = requests.post(url, data=json.dumps(json_req), headers=headers)
    subs_res = json.loads(response.json())

    dbg_log("get_subtitle_listing: [%s] sub_res len: %s" % (label, str(len(subs_res["Results"]))))

    if len(subs_res["Results"]) == 1:
        # Can't get better than that - the API found a single match (recall we looked for subtitle).
        list_item = ListItem(label="Hebrew", label2=subs_res["Results"][0]["SubtitleName"], thumbnailImage="he",
                             iconImage="5")
        list_item.setProperty("sync", "true")  # Because it's our best match
        url = "plugin://%s/?action=download&subsname=%s&id=%s" % (__scriptid__, subs_res["Results"][0]["SubtitleName"],
                                                                  subs_res["Results"][0]["Identifier"])
        addDirectoryItem(handle=int(sys.argv[1]), url=url, listitem=list_item, isFolder=False)
        return

    if label == "":
        return    # Can't do much from here (we probably relied solely on its IMDB id)

    # Going the trivial direct way yielded nothing so let's try to be a bit smarter
    title_name_guess = get_title_name(label)
    dbg_log("get_subtitle_listing: [%s] title_name_guess: %s" % (label, title_name_guess))

    json_req['request']['SearchPhrase'] = title_name_guess
    json_req['request']['SearchType'] = "FilmName"
    # dbg_log("get_subtitle_listing: "+str(json_req))

    # xbmc.sleep(500)
    response = requests.post(url, data=json.dumps(json_req), headers=headers)
    subs_res = json.loads(response.json())
    dbg_log("get_subtitle_listing: "+str(subs_res))
    # dbg_log("get_subtitle_listing: "+str(subs_res['ErrorMessage']).encode('cp1255'))

    if not subs_res["Results"]:
        return

    dbg_log("get_subtitle_listing: final sub_res len: " + str(len(subs_res["Results"])))

    # For best UX, we try to list the subtitles from top to bottom while giving our best heuristically-obtained
    # indication of how likely this subtitle to be the one we're looking for
    best_score = get_title_match_score(label, label)
    subs_list = []
    for subs_item in subs_res["Results"]:
        score = get_title_match_score(label, subs_item["SubtitleName"])
        rating = (score / float(best_score)) * 5
        rating = int(rating)    # Only full match will get 5 stars (that's why we don't round it)
        subs_list.append((subs_item["SubtitleName"], subs_item["Identifier"], rating, score))
    subs_list_score = sorted(subs_list, key=lambda name: name[3], reverse=True)
    for subs_elm in subs_list_score:
        list_item = ListItem(label="Hebrew", label2=subs_elm[0], thumbnailImage="he", iconImage=str(subs_elm[2]))
        # this url stuff does all the magic of invoking upon clicking a subtitle its actual text stream
        url = "plugin://%s/?action=download&subsname=%s&id=%s" % (__scriptid__, subs_elm[0], subs_elm[1])
        addDirectoryItem(handle=int(sys.argv[1]), url=url, listitem=list_item, isFolder=False)


def dbg_log(msg):
    log((u"##**## [%s] %s" % ("ScrewZira Subs", msg,)).encode('utf-8'), level=xbmc.LOGDEBUG)


if not exists(__temp__):
    mkdirs(__temp__)

action = None
if len(sys.argv) >= 2:
    params = get_params(sys.argv[2])
    action = get_param("action", params)

dbg_log("Version:%s" % __version__)
dbg_log("Action:%s" % action)

if action == 'search':
    item = {}

    dbg_log("isPlaying:%s" % Player().isPlaying())
    if Player().isPlaying():
        item['year'] = getInfoLabel("VideoPlayer.Year")  # Year

        item['season'] = str(getInfoLabel("VideoPlayer.Season"))  # Season
        if item['season'] == '' or item['season'] < 1:
            item['season'] = 0
        item['episode'] = str(getInfoLabel("VideoPlayer.Episode"))  # Episode
        if item['episode'] == '' or item['episode'] < 1:
            item['episode'] = 0

        if item['episode'] == 0:
            item['title'] = normalize_string(getInfoLabel("VideoPlayer.Title"))  # No original title, get just Title
        else:
            item['title'] = normalize_string(getInfoLabel("VideoPlayer.TVShowTitle"))  # Show
        if item['title'] == "":
            item['title'] = normalize_string(getInfoLabel("VideoPlayer.OriginalTitle"))  # Try to get original title
        item['file_original_path'] = unquote(unicode(Player().getPlayingFile(), 'utf-8'))  # Full path of a playing file
        item['file_original_path'] = item['file_original_path'].split("?")
        # CHECKME: Who wrote it assumed all extensions are 3 chars but what about .mpeg and such???
        #          I think path.splitext is the full-proof way to do it...
        item['file_original_path'] = path.basename(item['file_original_path'][0])[:-4]
        if item['season'] == 0:
            # When Kodi doesn't help us too much with the info, we must obtain ourselves the show info (if it's a show)
            # as ScrewZira API requires it so I'm relying here strictly on the global convention of "show.sNNeKK.mkv"
            # listing (which works in most cases).
            re_res = search(r"s(\d+)e(\d+)", item['title'].lower())
            if re_res:
                re_groups = re_res.groups()
                if len(re_groups) == 2:
                    try:
                        season_num = int(re_groups[0])
                        episode_num = int(re_groups[1])
                    except:
                        pass
                if episode_num:
                    item['season'] = season_num
                    item['episode'] = episode_num
        # If our title is simply the full file name then trim its extension
        if len(item['title']) > len(item['file_original_path']):
            item['title'] = path.splitext(item['title'])[0]
    else:  # Take item params from window when kodi is not playing
        labelIMDB = getInfoLabel("ListItem.IMDBNumber")
        item['year'] = getInfoLabel("ListItem.Year")
        item['season'] = getInfoLabel("ListItem.Season")
        item['episode'] = getInfoLabel("ListItem.Episode")
        item['file_original_path'] = ""
        labelType = getInfoLabel("ListItem.DBTYPE")  # movie/tvshow/season/episode
        isItMovie = labelType == 'movie' or getCondVisibility("Container.Content(movies)")
        isItEpisode = labelType == 'episode' or getCondVisibility("Container.Content(episodes)")

        if isItMovie:
            item['title'] = getInfoLabel("ListItem.OriginalTitle")
        elif isItEpisode:
            item['title'] = getInfoLabel("ListItem.TVShowTitle")
        else:
            item['title'] = "SearchFor..."  # In order to show "No Subtitles Found" result.

    dbg_log("item:%s" % item)
    imdb_id = 0
    try:
        if Player().isPlaying():  # Enable using subtitles search dialog when kodi is not playing
            playerid_query = '{"jsonrpc": "2.0", "method": "Player.GetActivePlayers", "id": 1}'
            playerid = json.loads(executeJSONRPC(playerid_query))['result'][0]['playerid']
            imdb_id_query = '{"jsonrpc": "2.0", "method": "Player.GetItem", "params": {"playerid": ' + str(
                playerid) + ', "properties": ["imdbnumber"]}, "id": 1}'
            imdb_id = json.loads(executeJSONRPC(imdb_id_query))['result']['item']['imdbnumber']
            dbg_log("imdb JSONPC:%s" % imdb_id)
        else:
            if labelIMDB:
                imdb_id = labelIMDB
            else:
                if isItMovie:
                    imdb_id = "ThisIsMovie"  # Search the movie by item['title'] for imdb_id
                elif isItEpisode:
                    imdb_id = "ThisIsEpisode"  # Search by item['title'] for tvdb_id
                else:
                    imdb_id = "tt0"  # In order to show "No Subtitles Found" result => Doesn't recognize movie/episode
    except:
        pass

    get_subtitle_listing(item['title'], imdb_id, item['season'], item['episode'])

    # Search Local File
    endOfDirectory(int(sys.argv[1]))
    if __addon__.getSetting("Debug") == "true":
        if imdb_id[:2] == "tt":
            Dialog().ok("Debug " + __version__, str(item), "imdb: " + str(imdb_id))
        else:
            Dialog().ok("Debug " + __version__, str(item), "NO IDS, use fname "+item["title"])

elif action == 'manualsearch':
    search_string = get_param("searchstring", params)
    get_subtitle_listing(search_string)
    endOfDirectory(int(sys.argv[1]))

elif action == 'download':
    id = get_param("id", params)
    subs_name = get_param("subsname", params)
    # dbg_log("Download ID:%s name:%s" % (id, subs_name))
    subs_path = download(id, subs_name)
    # We always fetch a single subtitle file (ScrewZira API)
    if subs_path != "":
        list_item = ListItem(label=subs_name)
        addDirectoryItem(handle=int(sys.argv[1]), url=subs_path, listitem=list_item, isFolder=False)
        endOfDirectory(int(sys.argv[1]))

elif action == 'clean':
    try:
        rmtree(__temp__)
    except:
        pass
    executebuiltin((u'Notification(%s,%s)' % (__scriptname__, __language__(32004))).encode('utf-8'))
